const assert = require('assert');
const signer = require('./signer');
const httpModule = require('http');
const httpsModule = require('https');
const axios = require('axios');
const helper = require('./helper');

class Client {

  constructor(config = {}) {
    
    if (config.endpoint) {
      if (!config.endpoint.startsWith('https://') &&
        !config.endpoint.startsWith('http://')) {
        throw new Error(`"config.endpoint" must starts with 'https://' or 'http://'.`);
      }
    }

    if (config.accessKeyId || config.accessKeySecret) {
      assert(config.accessKeyId, 'must pass "config.accessKeyId"');
      assert(config.accessKeySecret, 'must pass "config.accessKeySecret"');
    }

    if (config.securityToken) {
      assert(config.securityToken, 'must pass "config.securityToken"');
    }

    if (config.privateKey) {
      assert(config.privateKey, 'must pass "config.privateKey"');
    }

    if (config.publicKey) {
      assert(config.publicKey, 'must pass "config.publicKey"');
    }

    this.endpoint = config.endpoint;
    this.defaultEndpoint = "https://us-central1-bilipay.cloudfunctions.net";
    this.accessKeyId = config.accessKeyId;
    this.accessKeySecret = config.accessKeySecret;
    this.securityToken = config.securityToken;
    this.privateKey = config.privateKey;
    this.publicKey = config.publicKey;
    this.apiVersion = config.apiVersion || helper.PACKAGE_VERSION;
    this.signatureVersion = config.signatureVersion || helper.PACKAGE_VERSION;
    this.timeout = config.timeout || 30000;
    this.opts = config.opts || {};
    this.canonicalize = config.canonicalize || function(data){
      return helper.Canonicalize(data)
      // return JSON.stringify(data)
    };
  }

  _getUrl(uri, url){
    uri = url || uri
    if (!uri.startsWith("https://") && 
      !uri.startsWith("http://")) {
      var endpoint = this.endpoint || this.defaultEndpoint;
      return `${endpoint}/${/^\//i.test(uri) ? uri.substr(1) : uri}`;
    }
    return uri
  }

  _getKeepAliveAgent(endpoint){
    return endpoint.startsWith('https://')
      ? new httpsModule.Agent({
        keepAlive: true,
        keepAliveMsecs: this.timeout
      }) : new httpModule.Agent({
        keepAlive: true,
        keepAliveMsecs: this.timeout
      });
  }

  async request(url, opts = {}) {

    url = this._getUrl(url, opts.url)

    const method = (opts.method || 'GET').toUpperCase();

    if (opts && !opts.agent) {
      opts.agent = this._getKeepAliveAgent(url);
    }

    if (opts && !opts.timeout) {
      opts.timeout = this.timeout
    }

    // 签名
    const sig = new signer.Signer();
    //Set the AK/SK to sign and authenticate the request.
    sig.Key = this.accessKeyId;
    sig.Secret = this.accessKeySecret;
    sig.Privatekey = this.privateKey;
    sig.PublicKey = this.publicKey;
    sig.SecurityToken = this.securityToken;
    const r = new signer.HttpRequest(method, url);

    r.headers = {
      "x-core-nonce": helper.makeNonce(),
      "x-core-version": this.apiVersion,
      "x-core-signature-version": this.signatureVersion,
      "x-core-client": helper.DEFAULT_CLIENT,
      "x-core-ua": helper.DEFAULT_UA,
      "x-core-path": r.uri
    }

    // if(r.headers["x-forwarded-host"]){
    //   headers['host'] = r.headers["x-forwarded-host"]
    // }
    // if(r.headers["x-real-host"]){
    //   headers['host'] = r.headers["x-real-host"]
    // }

    if(r.method === 'PUT' || r.method === 'PATCH' || r.method === 'POST'){
      //Add header parameters, for example, x-domain-id for invoking a global service and x-project-id for invoking a project-level service.
      r.headers["Content-Type"] = (opts.headers && opts.headers["Content-Type"]) || "application/x-www-form-urlencoded";
      //Add a body if you have specified the PUT or POST method. Special characters, such as the double quotation mark ("), contained in the body must be escaped.
      r.body = this.canonicalize ? this.canonicalize(opts.data) : opts.data
    }

    r.query = Object.assign(r.query, opts.params)

    let opt = {};
    if(opts.web3 && opts.web3.signature && opts.web3.address){
      r.headers["x-core-authtype"] = "web3-sign-v1";
      r.headers["x-web3-signature"] = opts.web3.signature;
      r.headers["x-web3-address"] = opts.web3.address;
      r.headers["Content-Type"] = "application/json";
      r.headers["x-web3-timestamp"] = opts.web3.timestamp || null;
      opt = sig.Web3SignV1(r);
      delete opts.web3
    }else
    if(this.privateKey){
      r.headers["x-core-authtype"] = "web3";
      opt = sig.Web3Sign(r);
    }else
    if(this.accessKeyId && this.accessKeySecret){
      r.headers["x-core-authtype"] = "signature";
      opt = sig.Sign(r);
    }else
    if(this.securityToken){
      r.headers["x-core-authtype"] = "token";
      opt = sig.Token(r);
    }

    const entry = {
      url: url,
      request: null,
      response: null
    };
    
    const instance = axios.create();
    instance.interceptors.request.use(function (config) {
      return new Promise((resolve)=>{
        resolve(config)
      });
    }, function (error) {
      return Promise.reject(error);
    });

    return instance({
      url: url,
      ...this.opts,
      ...opts,
      ...opt,
      headers: Object.assign({
        "x-real-host": r.headers.host,
      }, this.opts.headers, opts.headers, opt.headers)
    })
    .then((response)=>{
      entry.request = {
        headers: response.request._header
      };
      entry.response = {
        statusCode: response.status,
        headers: response.headers
      };
      if(response.status === 200){
        return response.data
      }
      return {
        code: response.status,
        message: `request url: ${url} failed`,
        name: response.statusText,
        url: url,
        entry: entry
      }
    })
    .catch((err)=>{
      if(err.response){
        entry.request = {
          headers: err.response.request._header
        };
        entry.response = {
          statusCode: err.response.status,
          headers: err.response.headers
        }; 
        return {
          code: err.response.status,
          // message: `Request ${url} failed with status code ${err.response.status}`,
          message: err.message,
          name: err.code,
          url: url,
          entry: entry
        }
      }
      return {
        code: 443,
        message: err.message,
        name: err.code,
        url: url,
        entry: entry
      }
    })
  }

  async web3SignRequestBodyString(body = {}){
    const bodys = {
      method: "POST",
      nonce: helper.makeNonce(),
      body: body
    }
  }

  async post(url, data = {}, opts = {}) {
    return this.request(url, {
      method: "POST",
      data: data,
      ...opts
    })
  }

  async patch(url, data = {}, opts = {}) {
    return this.request(url, {
      method: "PATCH",
      data: data,
      ...opts
    })
  }

  async put(url, data = {}, opts = {}) {
    return this.request(url, {
      method: "PUT",
      data: data,
      ...opts
    })
  }

  async get(url, params = {}, opts = {}) {
    return this.request(url, {
      method: "GET",
      params: params,
      ...opts
    })
  }
  
  async delete(url, params = {}, opts = {}) {
    return this.request(url, {
      method: "DELETE",
      params: params,
      ...opts
    })
  }
}

module.exports = Client