'use strict';

const os = require('os');
const crypto = require('crypto');
const pkg = require('../package.json');

exports.DEFAULT_UA = `Bilipay (${os.platform()}; ${os.arch()}) ` +
  `Node.js/${process.version} Core/${pkg.version}`;
exports.DEFAULT_CLIENT = `Node.js(${process.version}), ${pkg.name}: ${pkg.version}`;
exports.PACKAGE_VERSION = `${pkg.version}`;


/**
 * Generate a date string
 * @return {String} a nonce string.
 */
exports.makeDate = (function () {
  return function () {
    var date = new Date();
    var YYYY = date.getUTCFullYear();
    var MM = twoChar(date.getUTCMonth() + 1);
    var DD = twoChar(date.getUTCDate());
    var HH = twoChar(date.getUTCHours());
    var mm = twoChar(date.getUTCMinutes());
    var ss = twoChar(date.getUTCSeconds());
    return `${YYYY}-${MM}-${DD}T${HH}:${mm}:${ss}Z`;
  }
}());

/**
 * Generate a nonce string
 * @return {String} a nonce string.
 */
exports.makeNonce = (function () {
  var counter = 0;
  var last;
  const machine = os.hostname();
  const pid = process.pid;

  return function () {
    var val = Math.floor(Math.random() * 1000000000000);
    if (val === last) {
      counter++;
    } else {
      counter = 0;
    }

    last = val;

    var uid = `${machine}${pid}${val}${counter}`;
    return exports.md5(uid, 'hex');
  };
}());

/**
 * Generate a haser with specfied algorithm
 * @param {String} algorithm can be md5, etc.
 * @return {Function} a haser with specfied algorithm
 */
 exports.makeHasher = function (algorithm) {
  return function (data, encoding) {
    var shasum = crypto.createHash(algorithm);
    shasum.update(data);
    return shasum.digest(encoding);
  };
};

/**
 * Get md5 hash digests of data
 * @param {String|Buffer} data data.
 * @param {String} encoding optional. can be 'hex', 'binary', 'base64'.
 * @return {String|Buffer} if no encoding is provided, a buffer is returned.
 */
exports.md5 = exports.makeHasher('md5');


/**
 * encode params data
 * @param {Object} params params.
 * @return {String|Buffer} if no encoding is provided, a buffer is returned.
 */
exports.Canonicalize = function(params) {
  var normalized = normalize(params);
  var fields = [];
  for (var i = 0; i < normalized.length; i++) {
    var [key, value] = normalized[i];
    fields.push(key + '=' + value);
  }
  return fields.join('&');
}

function encode(str) {
  var result = encodeURIComponent(str);

  return result.replace(/!/g, '%21')
    .replace(/'/g, '%27')
    .replace(/\(/g, '%28')
    .replace(/\)/g, '%29')
    .replace(/\*/g, '%2A');
}

function replaceRepeatList(target, key, repeat) {
  for (var i = 0; i < repeat.length; i++) {
    var item = repeat[i];

    if (Array.isArray(item)) {
      replaceRepeatList(target, `${key}.${i + 1}`, item);
    } else if (item && typeof item === 'object') {
      const keys = Object.keys(item);
      for (var j = 0; j < keys.length; j++) {
        if (Array.isArray(item[keys[j]])) {
          replaceRepeatList(target, `${key}.${i + 1}.${keys[j]}`, item[keys[j]]);
        } else {
          target[`${key}.${i + 1}.${keys[j]}`] = item[keys[j]];
        }
      }
    } else {
      target[`${key}.${i + 1}`] = item;
    }
  }
}

function flatParams(params) {
  var target = {};
  var keys = Object.keys(params);
  for (let i = 0; i < keys.length; i++) {
    var key = keys[i];
    var value = params[key];
    if (Array.isArray(value)) {
      replaceRepeatList(target, key, value);
    } else {
      target[key] = value;
    }
  }
  return target;
}

function normalize(params) {
  var list = [];
  var flated = flatParams(params);
  var keys = Object.keys(flated).sort();
  for (let i = 0; i < keys.length; i++) {
    var key = keys[i];
    var value = flated[key];
    list.push([encode(key), encode(value)]); //push []
  }
  return list;
}

function twoChar(num) {
  if (num < 10) {
    return '0' + num;
  }
  return '' + num;
}